# Copyright (c) OpenMMLab. All rights reserved.
from .generator_discriminator import (StyleGAN2Discriminator,
                                      StyleGANv2Generator)

__all__ = ['StyleGANv2Generator', 'StyleGAN2Discriminator']
